@@ -31,10 +31,7 @@
 #include <dbus/dbus-glib.h>
 #include <dbus/dbus-glib-lowlevel.h>
 
-#ifdef HAVE_POLKIT_GNOME
-#include <polkit-gnome/polkit-gnome.h>
-#endif
-
+#include "gsm-marshal.h"
 #include "gsm-consolekit.h"
 
 #define CK_NAME      "org.freedesktop.ConsoleKit"
@@ -64,6 +61,7 @@ enum {
 
 enum {
         REQUEST_COMPLETED = 0,
+        PRIVILEGES_COMPLETED,
         LAST_SIGNAL
 };
 
@@ -71,7 +69,13 @@ static guint signals[LAST_SIGNAL] = { 0 };
 
 static void     gsm_consolekit_class_init   (GsmConsolekitClass *klass);
 static void     gsm_consolekit_init         (GsmConsolekit      *ck);
-static void     gsm_consolekit_finalize     (GObject              *object);
+static void     gsm_consolekit_finalize     (GObject            *object);
+
+static void     gsm_consolekit_free_dbus    (GsmConsolekit      *manager);
+
+static DBusHandlerResult gsm_consolekit_dbus_filter (DBusConnection *connection,
+                                                     DBusMessage    *message,
+                                                     void           *user_data);
 
 static void     gsm_consolekit_on_name_owner_changed (DBusGProxy        *bus_proxy,
                                                       const char        *name,
@@ -133,9 +137,39 @@ gsm_consolekit_class_init (GsmConsolekitClass *manager_class)
                               G_TYPE_NONE,
                               1, G_TYPE_POINTER);
 
+        signals [PRIVILEGES_COMPLETED] =
+                g_signal_new ("privileges-completed",
+                              G_OBJECT_CLASS_TYPE (object_class),
+                              G_SIGNAL_RUN_LAST,
+                              G_STRUCT_OFFSET (GsmConsolekitClass, privileges_completed),
+                              NULL,
+                              NULL,
+                              gsm_marshal_VOID__BOOLEAN_BOOLEAN_POINTER,
+                              G_TYPE_NONE,
+                              3, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_POINTER);
+
         g_type_class_add_private (manager_class, sizeof (GsmConsolekitPrivate));
 }
 
+static DBusHandlerResult
+gsm_consolekit_dbus_filter (DBusConnection *connection,
+                            DBusMessage    *message,
+                            void           *user_data)
+{
+        GsmConsolekit *manager;
+
+        manager = GSM_CONSOLEKIT (user_data);
+
+        if (dbus_message_is_signal (message,
+                                    DBUS_INTERFACE_LOCAL, "Disconnected") &&
+            strcmp (dbus_message_get_path (message), DBUS_PATH_LOCAL) == 0) {
+                gsm_consolekit_free_dbus (manager);
+                return DBUS_HANDLER_RESULT_HANDLED;
+        }
+
+        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+}
+
 static gboolean
 gsm_consolekit_ensure_ck_connection (GsmConsolekit  *manager,
                                      GError        **error)
@@ -146,6 +180,8 @@ gsm_consolekit_ensure_ck_connection (GsmConsolekit  *manager,
         connection_error = NULL;
 
         if (manager->priv->dbus_connection == NULL) {
+                DBusConnection *connection;
+
                 manager->priv->dbus_connection = dbus_g_bus_get (DBUS_BUS_SYSTEM,
                                                                  &connection_error);
 
@@ -154,6 +190,12 @@ gsm_consolekit_ensure_ck_connection (GsmConsolekit  *manager,
                         is_connected = FALSE;
                         goto out;
                 }
+
+                connection = dbus_g_connection_get_connection (manager->priv->dbus_connection);
+                dbus_connection_set_exit_on_disconnect (connection, FALSE);
+                dbus_connection_add_filter (connection,
+                                            gsm_consolekit_dbus_filter,
+                                            manager, NULL);
         }
 
         if (manager->priv->bus_proxy == NULL) {
@@ -263,6 +305,31 @@ gsm_consolekit_init (GsmConsolekit *manager)
         }
 }
 
+static void
+gsm_consolekit_free_dbus (GsmConsolekit *manager)
+{
+        if (manager->priv->bus_proxy != NULL) {
+                g_object_unref (manager->priv->bus_proxy);
+                manager->priv->bus_proxy = NULL;
+        }
+
+        if (manager->priv->ck_proxy != NULL) {
+                g_object_unref (manager->priv->ck_proxy);
+                manager->priv->ck_proxy = NULL;
+        }
+
+        if (manager->priv->dbus_connection != NULL) {
+                DBusConnection *connection;
+                connection = dbus_g_connection_get_connection (manager->priv->dbus_connection);
+                dbus_connection_remove_filter (connection,
+                                               gsm_consolekit_dbus_filter,
+                                               manager);
+
+                dbus_g_connection_unref (manager->priv->dbus_connection);
+                manager->priv->dbus_connection = NULL;
+        }
+}
+
 static void
 gsm_consolekit_finalize (GObject *object)
 {
@@ -273,13 +340,7 @@ gsm_consolekit_finalize (GObject *object)
 
         parent_class = G_OBJECT_CLASS (gsm_consolekit_parent_class);
 
-        if (manager->priv->bus_proxy != NULL) {
-                g_object_unref (manager->priv->bus_proxy);
-        }
-
-        if (manager->priv->ck_proxy != NULL) {
-                g_object_unref (manager->priv->ck_proxy);
-        }
+        gsm_consolekit_free_dbus (manager);
 
         if (parent_class->finalize != NULL) {
                 parent_class->finalize (object);
@@ -356,16 +417,16 @@ try_system_restart (DBusGConnection *connection,
 
 static void
 emit_restart_complete (GsmConsolekit *manager,
-                       const char    *error_message)
+                       GError        *error)
 {
         GError *call_error;
 
         call_error = NULL;
 
-        if (error_message != NULL) {
+        if (error != NULL) {
                 call_error = g_error_new_literal (GSM_CONSOLEKIT_ERROR,
                                                   GSM_CONSOLEKIT_ERROR_RESTARTING,
-                                                  error_message);
+                                                  error->message);
         }
 
         g_signal_emit (G_OBJECT (manager),
@@ -400,178 +461,6 @@ emit_stop_complete (GsmConsolekit *manager,
         }
 }
 
-#ifdef HAVE_POLKIT_GNOME
-static void
-system_restart_auth_cb (PolKitAction  *action,
-                        gboolean       gained_privilege,
-                        GError        *error,
-                        GsmConsolekit *manager)
-{
-        GError  *local_error;
-        gboolean res;
-
-        if (!gained_privilege) {
-                if (error != NULL) {
-                        emit_restart_complete (manager, error->message);
-                }
-
-                return;
-        }
-
-        local_error = NULL;
-
-        res = try_system_restart (manager->priv->dbus_connection, &local_error);
-
-        if (!res) {
-                g_warning ("Unable to restart system: %s", local_error->message);
-                emit_restart_complete (manager, local_error->message);
-                g_error_free (local_error);
-
-                return;
-        }
-}
-
-static void
-system_stop_auth_cb (PolKitAction  *action,
-                     gboolean       gained_privilege,
-                     GError        *error,
-                     GsmConsolekit *manager)
-{
-        GError  *local_error;
-        gboolean res;
-
-        if (!gained_privilege) {
-                if (error != NULL) {
-                        emit_stop_complete (manager, error);
-                }
-
-                return;
-        }
-
-        local_error = NULL;
-
-        res = try_system_stop (manager->priv->dbus_connection, &local_error);
-
-        if (!res) {
-                g_warning ("Unable to stop system: %s", local_error->message);
-                emit_stop_complete (manager, local_error);
-                g_error_free (local_error);
-
-                return;
-        }
-}
-
-static PolKitAction *
-get_action_from_error (GError *error)
-{
-        PolKitAction *action;
-        char         *paction;
-
-        action = polkit_action_new ();
-
-        paction = NULL;
-        if (g_str_has_prefix (error->message, "Not privileged for action: ")) {
-                paction = g_strdup (error->message + strlen ("Not privileged for action: "));
-                if (paction != NULL) {
-                        char *p;
-
-                        /* after 0.2.10 the error also includes the PK results */
-                        p = strchr (paction, ' ');
-                        if (p != NULL) {
-                                *p = '\0';
-                        }
-                }
-        }
-
-        polkit_action_set_action_id (action, paction);
-
-        g_free (paction);
-
-        return action;
-}
-#endif /* HAVE_POLKIT_GNOME */
-
-static void
-request_restart_priv (GsmConsolekit *manager,
-                      GError        *error)
-{
-#ifdef HAVE_POLKIT_GNOME
-        PolKitAction *action;
-        pid_t         pid;
-        char         *error_message = NULL;
-        gboolean      res = FALSE;
-        guint         xid;
-        GError       *local_error;
-
-        action = get_action_from_error (error);
-
-        xid = 0;
-        pid = getpid ();
-
-        local_error = NULL;
-        res = polkit_gnome_auth_obtain (action,
-                                        xid,
-                                        pid,
-                                        (PolKitGnomeAuthCB) system_restart_auth_cb,
-                                        manager,
-                                        &local_error);
-
-        polkit_action_unref (action);
-
-        if (local_error != NULL) {
-                error_message = g_strdup (local_error->message);
-                g_error_free (local_error);
-        }
-
-        if (!res) {
-                emit_restart_complete (manager, error_message);
-                g_free (error_message);
-        }
-#else
-        g_assert_not_reached ();
-#endif /* HAVE POLKIT */
-}
-
-static void
-request_stop_priv (GsmConsolekit *manager,
-                   GError        *error)
-{
-#ifdef HAVE_POLKIT_GNOME
-        PolKitAction *action;
-        pid_t         pid;
-        gboolean      res = FALSE;
-        guint         xid;
-        GError       *local_error;
-
-        action = get_action_from_error (error);
-
-        xid = 0;
-        pid = getpid ();
-
-        local_error = NULL;
-        res = polkit_gnome_auth_obtain (action,
-                                        xid,
-                                        pid,
-                                        (PolKitGnomeAuthCB) system_stop_auth_cb,
-                                        manager,
-                                        &local_error);
-
-        polkit_action_unref (action);
-
-        if (!res) {
-                if (local_error != NULL) {
-                        g_warning ("Unable to obtain auth to stop system: %s",
-                                   local_error->message);
-
-                        emit_stop_complete (manager, local_error);
-                        g_error_free (local_error);
-                }
-        }
-#else
-        g_assert_not_reached ();
-#endif /* HAVE POLKIT */
-}
-
 void
 gsm_consolekit_attempt_restart (GsmConsolekit *manager)
 {
@@ -583,20 +472,24 @@ gsm_consolekit_attempt_restart (GsmConsolekit *manager)
         if (!gsm_consolekit_ensure_ck_connection (manager, &error)) {
                 g_warning ("Could not connect to ConsoleKit: %s",
                            error->message);
+                emit_restart_complete (manager, error);
                 g_error_free (error);
                 return;
         }
 
-        res = try_system_restart (manager->priv->dbus_connection, &error);
+        res = dbus_g_proxy_call_with_timeout (manager->priv->ck_proxy,
+                                              "Restart",
+                                              INT_MAX,
+                                              error,
+                                              G_TYPE_INVALID,
+                                              G_TYPE_INVALID);
 
         if (!res) {
-                if (dbus_g_error_has_name (error, "org.freedesktop.ConsoleKit.Manager.NotPrivileged")) {
-                        request_restart_priv (manager, error);
-                } else {
-                        emit_restart_complete (manager, error->message);
-                }
-
+                g_warning ("Unable to restart system: %s", error->message);
+                emit_restart_complete (manager, error);
                 g_error_free (error);
+        } else {
+                emit_restart_complete (manager, NULL);
         }
 }
 
@@ -611,21 +504,24 @@ gsm_consolekit_attempt_stop (GsmConsolekit *manager)
         if (!gsm_consolekit_ensure_ck_connection (manager, &error)) {
                 g_warning ("Could not connect to ConsoleKit: %s",
                            error->message);
+                emit_stop_complete (manager, error);
                 g_error_free (error);
                 return;
         }
 
-        res = try_system_stop (manager->priv->dbus_connection, &error);
+        res = dbus_g_proxy_call_with_timeout (manager->priv->ck_proxy,
+                                              "Stop",
+                                              INT_MAX,
+                                              error,
+                                              G_TYPE_INVALID,
+                                              G_TYPE_INVALID);
 
         if (!res) {
                 g_warning ("Unable to stop system: %s", error->message);
-                if (dbus_g_error_has_name (error, "org.freedesktop.ConsoleKit.Manager.NotPrivileged")) {
-                        request_stop_priv (manager, error);
-                } else {
-                        emit_stop_complete (manager, error);
-                }
-
+                emit_stop_complete (manager, error);
                 g_error_free (error);
+        } else {
+                emit_stop_complete (manager, NULL);
         }
 }
 
@@ -872,9 +768,19 @@ seat_can_activate_sessions (DBusConnection *connection,
 gboolean
 gsm_consolekit_can_switch_user (GsmConsolekit *manager)
 {
+        GError  *error;
         char    *seat_id;
         gboolean ret;
 
+        error = NULL;
+
+        if (!gsm_consolekit_ensure_ck_connection (manager, &error)) {
+                g_warning ("Could not connect to ConsoleKit: %s",
+                           error->message);
+                g_error_free (error);
+                return FALSE;
+        }
+
         seat_id = get_current_seat_id (dbus_g_connection_get_connection (manager->priv->dbus_connection));
         if (seat_id == NULL || seat_id[0] == '\0') {
                 g_debug ("seat id is not set; can't switch sessions");
@@ -888,107 +794,148 @@ gsm_consolekit_can_switch_user (GsmConsolekit *manager)
         return ret;
 }
 
-#ifdef HAVE_POLKIT_GNOME
-static gboolean
-gsm_consolekit_can_do_action (GsmConsolekit *manager,
-                              const char    *action_id)
+gboolean
+gsm_consolekit_get_restart_privileges (GsmConsolekit *manager)
 {
-        PolKitGnomeContext *gnome_context;
-        PolKitAction *action;
-        PolKitCaller *caller;
-        DBusError dbus_error;
-        PolKitError *error;
-        PolKitResult result;
+	g_signal_emit (G_OBJECT (manager),
+		       signals [PRIVILEGES_COMPLETED],
+		       0, TRUE, TRUE, NULL);
 
-        gnome_context = polkit_gnome_context_get (NULL);
+	return TRUE;
+}
 
-        if (gnome_context == NULL) {
-                return FALSE;
-        }
+gboolean
+gsm_consolekit_get_stop_privileges (GsmConsolekit *manager)
+{
+	g_signal_emit (G_OBJECT (manager),
+		       signals [PRIVILEGES_COMPLETED],
+		       0, TRUE, TRUE, NULL);
 
-        if (gnome_context->pk_tracker == NULL) {
-                return FALSE;
-        }
-
-        dbus_error_init (&dbus_error);
-        caller = polkit_tracker_get_caller_from_pid (gnome_context->pk_tracker,
-                                                     getpid (),
-                                                     &dbus_error);
-        dbus_error_free (&dbus_error);
-
-        if (caller == NULL) {
-                return FALSE;
-        }
-
-        action = polkit_action_new ();
-        if (!polkit_action_set_action_id (action, action_id)) {
-                polkit_action_unref (action);
-                polkit_caller_unref (caller);
-                return FALSE;
-        }
-
-        result = POLKIT_RESULT_UNKNOWN;
-        error = NULL;
-        result = polkit_context_is_caller_authorized (gnome_context->pk_context,
-                                                      action, caller, FALSE,
-                                                      &error);
-        if (polkit_error_is_set (error)) {
-                polkit_error_free (error);
-        }
-        polkit_action_unref (action);
-                polkit_caller_unref (caller);
-
-        return result != POLKIT_RESULT_NO && result != POLKIT_RESULT_UNKNOWN;
+	return TRUE;
 }
-#endif
 
 gboolean
 gsm_consolekit_can_restart (GsmConsolekit *manager)
 {
-#ifdef HAVE_POLKIT_GNOME
         gboolean res;
+	gboolean can_restart;
         GError  *error;
 
         error = NULL;
-        res = gsm_consolekit_ensure_ck_connection (manager, &error);
-        if (!res) {
+
+        if (!gsm_consolekit_ensure_ck_connection (manager, &error)) {
                 g_warning ("Could not connect to ConsoleKit: %s",
                            error->message);
                 g_error_free (error);
-                return FALSE;
+                return;
         }
 
-        return gsm_consolekit_can_do_action (manager, "org.freedesktop.consolekit.system.restart") ||
-               gsm_consolekit_can_do_action (manager, "org.freedesktop.consolekit.system.restart-multiple-users");
-#else
-        g_debug ("GsmConsolekit: built without PolicyKit-gnome support - cannot restart system");
-        return FALSE;
-#endif
+        res = dbus_g_proxy_call_with_timeout (manager->priv->ck_proxy,
+                                              "CanRestart",
+                                              INT_MAX,
+                                              error,
+                                              G_TYPE_INVALID,
+                                              G_TYPE_BOOLEAN, &can_restart,
+                                              G_TYPE_INVALID);
+
+	return can_restart;
 }
 
 gboolean
 gsm_consolekit_can_stop (GsmConsolekit *manager)
 {
-#ifdef HAVE_POLKIT_GNOME
         gboolean res;
+	gboolean can_stop;
         GError  *error;
+
         error = NULL;
-        res = gsm_consolekit_ensure_ck_connection (manager, &error);
-        if (!res) {
+
+        if (!gsm_consolekit_ensure_ck_connection (manager, &error)) {
                 g_warning ("Could not connect to ConsoleKit: %s",
                            error->message);
                 g_error_free (error);
-                return FALSE;
+                return;
         }
 
-        return gsm_consolekit_can_do_action (manager, "org.freedesktop.consolekit.system.stop") ||
-               gsm_consolekit_can_do_action (manager, "org.freedesktop.consolekit.system.stop-multiple-users");
-#else
-        g_debug ("GsmConsolekit: built without PolicyKit-gnome support - cannot stop system");
-        return FALSE;
-#endif
+        res = dbus_g_proxy_call_with_timeout (manager->priv->ck_proxy,
+                                              "CanStop",
+                                              INT_MAX,
+                                              error,
+                                              G_TYPE_INVALID,
+                                              G_TYPE_BOOLEAN, &can_stop,
+                                              G_TYPE_INVALID);
+
+	return can_stop;
 }
 
+gchar *
+gsm_consolekit_get_current_session_type (GsmConsolekit *manager)
+{
+        GError *gerror;
+	DBusConnection *connection;
+	DBusError error;
+	DBusMessage *message = NULL;
+	DBusMessage *reply = NULL;
+	gchar *session_id;
+	gchar *ret;
+	DBusMessageIter iter;
+	const char *value;
+
+	session_id = NULL;
+	ret = NULL;
+        gerror = NULL;
+
+        if (!gsm_consolekit_ensure_ck_connection (manager, &gerror)) {
+                g_warning ("Could not connect to ConsoleKit: %s",
+                           gerror->message);
+                g_error_free (gerror);
+		goto out;
+        }
+
+	connection = dbus_g_connection_get_connection (manager->priv->dbus_connection);
+	if (!get_current_session_id (connection, &session_id)) {
+		goto out;
+	}
+
+	dbus_error_init (&error);
+	message = dbus_message_new_method_call (CK_NAME,
+						session_id,
+						CK_SESSION_INTERFACE,
+						"GetSessionType");
+	if (message == NULL) {
+		goto out;
+	}
+
+	reply = dbus_connection_send_with_reply_and_block (connection,
+							   message,
+							   -1,
+							   &error);
+
+	if (reply == NULL) {
+		if (dbus_error_is_set (&error)) {
+			g_warning ("Unable to determine session type: %s", error.message);
+			dbus_error_free (&error);
+		}
+		goto out;
+	}
+
+	dbus_message_iter_init (reply, &iter);
+	dbus_message_iter_get_basic (&iter, &value);
+	ret = g_strdup (value);
+
+out:
+	if (message != NULL) {
+		dbus_message_unref (message);
+	}
+	if (reply != NULL) {
+		dbus_message_unref (reply);
+	}
+	g_free (session_id);
+
+	return ret;
+}
+
+
 GsmConsolekit *
 gsm_get_consolekit (void)
 {
